@@ -25,7 +25,7 @@ struct AnyRegister {
     Code code_;
 
   public:
-    AnyRegister() = default;
+    AnyRegister() : code_(Invalid) {}
 
     explicit AnyRegister(Register gpr) {
         code_ = gpr.code();
@@ -40,6 +40,7 @@ struct AnyRegister {
         return r;
     }
     bool isFloat() const {
+        MOZ_ASSERT(isValid());
         return code_ >= Registers::Total;
     }
     Register gpr() const {
@@ -51,15 +52,18 @@ struct AnyRegister {
         return FloatRegister::FromCode(code_ - Registers::Total);
     }
     bool operator ==(AnyRegister other) const {
+        // We don't need the operands to be valid to test for equality.
         return code_ == other.code_;
     }
     bool operator !=(AnyRegister other) const {
+        // We don't need the operands to be valid to test for equality.
         return code_ != other.code_;
     }
     const char* name() const {
         return isFloat() ? fpu().name() : gpr().name();
     }
     Code code() const {
+        MOZ_ASSERT(isValid());
         return code_;
     }
     bool volatile_() const {
@@ -99,7 +103,9 @@ struct AnyRegister {
             return true;
         return false;
     }
-
+    bool isValid() const {
+        return code_ != Invalid;
+    }
 };
 
 // Registers to hold a boxed value. Uses one register on 64 bit
@@ -124,13 +130,13 @@ class ValueOperand
     bool aliases(Register reg) const {
         return type_ == reg || payload_ == reg;
     }
-    Register scratchReg() const {
+    Register payloadOrValueReg() const {
         return payloadReg();
     }
-    bool operator==(const ValueOperand& o) const {
+    constexpr bool operator==(const ValueOperand& o) const {
         return type_ == o.type_ && payload_ == o.payload_;
     }
-    bool operator!=(const ValueOperand& o) const {
+    constexpr bool operator!=(const ValueOperand& o) const {
         return !(*this == o);
     }
 
@@ -148,17 +154,21 @@ class ValueOperand
     bool aliases(Register reg) const {
         return value_ == reg;
     }
-    Register scratchReg() const {
+    Register payloadOrValueReg() const {
         return valueReg();
     }
-    bool operator==(const ValueOperand& o) const {
+    constexpr bool operator==(const ValueOperand& o) const {
         return value_ == o.value_;
     }
-    bool operator!=(const ValueOperand& o) const {
+    constexpr bool operator!=(const ValueOperand& o) const {
         return !(*this == o);
     }
 #endif
 
+    Register scratchReg() const {
+        return payloadOrValueReg();
+    }
+
     ValueOperand() = default;
 };
 
@@ -169,7 +179,7 @@ class TypedOrValueRegister
     MIRType type_;
 
     union U {
-        AnyRegister typed;
+        AnyRegister::Code typed;
         ValueOperand value;
     } data;
 
@@ -180,7 +190,7 @@ class TypedOrValueRegister
     TypedOrValueRegister(MIRType type, AnyRegister reg)
       : type_(type)
     {
-        data.typed = reg;
+        data.typed = reg.code();
     }
 
     MOZ_IMPLICIT TypedOrValueRegister(ValueOperand value)
@@ -203,7 +213,7 @@ class TypedOrValueRegister
 
     AnyRegister typedReg() const {
         MOZ_ASSERT(hasTyped());
-        return data.typed;
+        return AnyRegister::FromCode(data.typed);
     }
 
     ValueOperand valueReg() const {
@@ -226,13 +236,13 @@ class ConstantOrRegister
 
     // Space to hold either a Value or a TypedOrValueRegister.
     union U {
-        Value constant;
+        JS::UninitializedValue constant;
         TypedOrValueRegister reg;
     } data;
 
-    const Value& dataValue() const {
+    Value dataValue() const {
         MOZ_ASSERT(constant());
-        return data.constant;
+        return data.constant.asValueRef();
     }
     void setDataValue(const Value& value) {
         MOZ_ASSERT(constant());
@@ -268,7 +278,7 @@ class ConstantOrRegister
         return constant_;
     }
 
-    const Value& value() const {
+    Value value() const {
         return dataValue();
     }
 
@@ -277,41 +287,6 @@ class ConstantOrRegister
     }
 };
 
-struct RegisterOrInt32Constant {
-    bool isRegister_;
-    union {
-        Register reg_;
-        int32_t constant_;
-    };
-
-    explicit RegisterOrInt32Constant(Register reg)
-      : isRegister_(true), reg_(reg)
-    { }
-
-    explicit RegisterOrInt32Constant(int32_t index)
-      : isRegister_(false), constant_(index)
-    { }
-
-    inline void bumpConstant(int diff) {
-        MOZ_ASSERT(!isRegister_);
-        constant_ += diff;
-    }
-    inline Register reg() const {
-        MOZ_ASSERT(isRegister_);
-        return reg_;
-    }
-    inline int32_t constant() const {
-        MOZ_ASSERT(!isRegister_);
-        return constant_;
-    }
-    inline bool isRegister() const {
-        return isRegister_;
-    }
-    inline bool isConstant() const {
-        return !isRegister_;
-    }
-};
-
 template <typename T>
 class TypedRegisterSet
 {
@@ -392,22 +367,22 @@ class TypedRegisterSet
         bits_ &= ~reg.alignedOrDominatedAliasedSet();
     }
 
-    T getAny() const {
-        // The choice of first or last here is mostly arbitrary, as they are
-        // about the same speed on popular architectures. We choose first, as
-        // it has the advantage of using the "lower" registers more often. These
-        // registers are sometimes more efficient (e.g. optimized encodings for
-        // EAX on x86).
-        return getFirst();
+    static constexpr RegTypeName DefaultType = RegType::DefaultType;
+
+    template <RegTypeName Name>
+    SetType allLive() const {
+        return T::template LiveAsIndexableSet<Name>(bits_);
     }
-    T getFirst() const {
-        MOZ_ASSERT(!empty());
-        return T::FromCode(T::FirstBit(bits_));
+    template <RegTypeName Name>
+    SetType allAllocatable() const {
+        return T::template AllocatableAsIndexableSet<Name>(bits_);
     }
-    T getLast() const {
-        MOZ_ASSERT(!empty());
-        int ireg = T::LastBit(bits_);
-        return T::FromCode(ireg);
+
+    static RegType FirstRegister(SetType set) {
+        return RegType::FromCode(RegType::FirstBit(set));
+    }
+    static RegType LastRegister(SetType set) {
+        return RegType::FromCode(RegType::LastBit(set));
     }
 
     SetType bits() const {
@@ -481,6 +456,9 @@ class RegisterSet {
     bool emptyFloat() const {
         return fpu_.empty();
     }
+
+    static constexpr RegTypeName DefaultType = RegTypeName::GPR;
+
     constexpr GeneralRegisterSet gprs() const {
         return gpr_;
     }
@@ -496,7 +474,6 @@ class RegisterSet {
     bool operator ==(const RegisterSet& other) const {
         return other.gpr_ == gpr_ && other.fpu_ == fpu_;
     }
-
 };
 
 // There are 2 use cases for register sets:
@@ -540,6 +517,9 @@ class LiveSet;
 // Base accessors classes have the minimal set of raw methods to manipulate the register set
 // given as parameter in a consistent manner.  These methods are:
 //
+//    - all<Type>: Returns a bit-set of all the register of a specific type
+//      which are present.
+//
 //    - has: Returns if all the bits needed to take a register are present.
 //
 //    - takeUnchecked: Subtracts the bits used to represent the register in the
@@ -573,6 +553,11 @@ class AllocatableSetAccessors
   protected:
     RegSet set_;
 
+    template <RegTypeName Name>
+    SetType all() const {
+        return set_.template allAllocatable<Name>();
+    }
+
   public:
     AllocatableSetAccessors() : set_() {}
     explicit constexpr AllocatableSetAccessors(SetType set) : set_(set) {}
@@ -582,6 +567,11 @@ class AllocatableSetAccessors
         return set_.hasAllocatable(reg);
     }
 
+    template <RegTypeName Name>
+    bool hasAny(RegType reg) const {
+        return all<Name>() != 0;
+    }
+
     void addUnchecked(RegType reg) {
         set_.addAllocatable(reg);
     }
@@ -603,6 +593,15 @@ class AllocatableSetAccessors<RegisterSet>
   protected:
     RegisterSet set_;
 
+    template <RegTypeName Name>
+    GeneralRegisterSet::SetType allGpr() const {
+        return set_.gprs().allAllocatable<Name>();
+    }
+    template <RegTypeName Name>
+    FloatRegisterSet::SetType allFpu() const {
+        return set_.fpus().allAllocatable<Name>();
+    }
+
   public:
     AllocatableSetAccessors() : set_() {}
     explicit constexpr AllocatableSetAccessors(SetType) = delete;
@@ -655,6 +654,11 @@ class LiveSetAccessors
   protected:
     RegSet set_;
 
+    template <RegTypeName Name>
+    SetType all() const {
+        return set_.template allLive<Name>();
+    }
+
   public:
     LiveSetAccessors() : set_() {}
     explicit constexpr LiveSetAccessors(SetType set) : set_(set) {}
@@ -685,6 +689,15 @@ class LiveSetAccessors<RegisterSet>
   protected:
     RegisterSet set_;
 
+    template <RegTypeName Name>
+    GeneralRegisterSet::SetType allGpr() const {
+        return set_.gprs().allLive<Name>();
+    }
+    template <RegTypeName Name>
+    FloatRegisterSet::SetType allFpu() const {
+        return set_.fpus().allLive<Name>();
+    }
+
   public:
     LiveSetAccessors() : set_() {}
     explicit constexpr LiveSetAccessors(SetType) = delete;
@@ -751,47 +764,68 @@ class SpecializedRegSet : public Accessors
         takeUnchecked(reg);
     }
 
-    RegType getAny() const {
-        return this->Parent::set_.getAny();
-    }
-    RegType getFirst() const {
-        return this->Parent::set_.getFirst();
-    }
-    RegType getLast() const {
-        return this->Parent::set_.getLast();
+    template <RegTypeName Name>
+    bool hasAny() const {
+        return Parent::template all<Name>() != 0;
     }
 
+    template <RegTypeName Name = RegSet::DefaultType>
+    RegType getFirst() const {
+        SetType set = Parent::template all<Name>();
+        MOZ_ASSERT(set);
+        return RegSet::FirstRegister(set);
+    }
+    template <RegTypeName Name = RegSet::DefaultType>
+    RegType getLast() const {
+        SetType set = Parent::template all<Name>();
+        MOZ_ASSERT(set);
+        return RegSet::LastRegister(set);
+    }
+    template <RegTypeName Name = RegSet::DefaultType>
+    RegType getAny() const {
+        // The choice of first or last here is mostly arbitrary, as they are
+        // about the same speed on popular architectures. We choose first, as
+        // it has the advantage of using the "lower" registers more often. These
+        // registers are sometimes more efficient (e.g. optimized encodings for
+        // EAX on x86).
+        return getFirst<Name>();
+    }
+
+    template <RegTypeName Name = RegSet::DefaultType>
     RegType getAnyExcluding(RegType preclude) {
         if (!has(preclude))
-            return getAny();
+            return getAny<Name>();
 
         take(preclude);
-        RegType result = getAny();
+        RegType result = getAny<Name>();
         add(preclude);
         return result;
     }
 
+    template <RegTypeName Name = RegSet::DefaultType>
     RegType takeAny() {
-        RegType reg = getAny();
+        RegType reg = getAny<Name>();
         take(reg);
         return reg;
     }
+    template <RegTypeName Name = RegSet::DefaultType>
     RegType takeFirst() {
-        RegType reg = getFirst();
+        RegType reg = getFirst<Name>();
         take(reg);
         return reg;
     }
+    template <RegTypeName Name = RegSet::DefaultType>
     RegType takeLast() {
-        RegType reg = getLast();
+        RegType reg = getLast<Name>();
         take(reg);
         return reg;
     }
 
     ValueOperand takeAnyValue() {
 #if defined(JS_NUNBOX32)
-        return ValueOperand(takeAny(), takeAny());
+        return ValueOperand(takeAny<RegTypeName::GPR>(), takeAny<RegTypeName::GPR>());
 #elif defined(JS_PUNBOX64)
-        return ValueOperand(takeAny());
+        return ValueOperand(takeAny<RegTypeName::GPR>());
 #else
 #error "Bad architecture"
 #endif
@@ -805,8 +839,9 @@ class SpecializedRegSet : public Accessors
 #endif
     }
 
+    template <RegTypeName Name = RegSet::DefaultType>
     RegType takeAnyExcluding(RegType preclude) {
-        RegType reg = getAnyExcluding(preclude);
+        RegType reg = getAnyExcluding<Name>(preclude);
         take(reg);
         return reg;
     }
@@ -841,12 +876,17 @@ class SpecializedRegSet<Accessors, RegisterSet> : public Accessors
         return this->Parent::set_.emptyFloat();
     }
 
-
     using Parent::has;
     bool has(AnyRegister reg) const {
         return reg.isFloat() ? has(reg.fpu()) : has(reg.gpr());
     }
 
+    template <RegTypeName Name>
+    bool hasAny() const {
+        if (Name == RegTypeName::GPR)
+            return Parent::template allGpr<RegTypeName::GPR>() != 0;
+        return Parent::template allFpu<Name>() != 0;
+    }
 
     using Parent::addUnchecked;
     void addUnchecked(AnyRegister reg) {
@@ -895,10 +935,15 @@ class SpecializedRegSet<Accessors, RegisterSet> : public Accessors
     }
 
     Register getAnyGeneral() const {
-        return this->Parent::set_.gprs().getAny();
+        GeneralRegisterSet::SetType set = Parent::template allGpr<RegTypeName::GPR>();
+        MOZ_ASSERT(set);
+        return GeneralRegisterSet::FirstRegister(set);
     }
+    template <RegTypeName Name = RegTypeName::Float64>
     FloatRegister getAnyFloat() const {
-        return this->Parent::set_.fpus().getAny();
+        FloatRegisterSet::SetType set = Parent::template allFpu<Name>();
+        MOZ_ASSERT(set);
+        return FloatRegisterSet::FirstRegister(set);
     }
 
     Register takeAnyGeneral() {
@@ -906,8 +951,9 @@ class SpecializedRegSet<Accessors, RegisterSet> : public Accessors
         take(reg);
         return reg;
     }
+    template <RegTypeName Name = RegTypeName::Float64>
     FloatRegister takeAnyFloat() {
-        FloatRegister reg = getAnyFloat();
+        FloatRegister reg = getAnyFloat<Name>();
         take(reg);
         return reg;
     }
@@ -959,6 +1005,19 @@ class CommonRegSet : public SpecializedRegSet<Accessors, Set>
 #error "Bad architecture"
 #endif
     }
+
+    using Parent::addUnchecked;
+    void addUnchecked(ValueOperand value) {
+#if defined(JS_NUNBOX32)
+        addUnchecked(value.payloadReg());
+        addUnchecked(value.typeReg());
+#elif defined(JS_PUNBOX64)
+        addUnchecked(value.valueReg());
+#else
+#error "Bad architecture"
+#endif
+    }
+
     void add(TypedOrValueRegister reg) {
         if (reg.hasValue())
             add(reg.valueReg());
@@ -1105,11 +1164,11 @@ class TypedRegisterIterator
         return !regset_.empty();
     }
     TypedRegisterIterator<T>& operator ++() {
-        regset_.takeAny();
+        regset_.template takeAny<RegTypeName::Any>();
         return *this;
     }
     T operator*() const {
-        return regset_.getAny();
+        return regset_.template getAny<RegTypeName::Any>();
     }
 };
 
@@ -1132,11 +1191,11 @@ class TypedRegisterBackwardIterator
         return !regset_.empty();
     }
     TypedRegisterBackwardIterator<T>& operator ++() {
-        regset_.takeLast();
+        regset_.template takeLast<RegTypeName::Any>();
         return *this;
     }
     T operator*() const {
-        return regset_.getLast();
+        return regset_.template getLast<RegTypeName::Any>();
     }
 };
 
@@ -1158,11 +1217,11 @@ class TypedRegisterForwardIterator
         return !regset_.empty();
     }
     TypedRegisterForwardIterator<T>& operator ++() {
-        regset_.takeFirst();
+        regset_.template takeFirst<RegTypeName::Any>();
         return *this;
     }
     T operator*() const {
-        return regset_.getFirst();
+        return regset_.template getFirst<RegTypeName::Any>();
     }
 };
 
@@ -1220,7 +1279,8 @@ class ABIArg
         GPR_PAIR,
 #endif
         FPU,
-        Stack
+        Stack,
+        Uninitialized = -1
     };
 
   private:
@@ -1232,7 +1292,7 @@ class ABIArg
     } u;
 
   public:
-    ABIArg() : kind_(Kind(-1)) { u.offset_ = -1; }
+    ABIArg() : kind_(Uninitialized) { u.offset_ = -1; }
     explicit ABIArg(Register gpr) : kind_(GPR) { u.gpr_ = gpr.code(); }
     explicit ABIArg(Register gprLow, Register gprHigh)
     {
@@ -1248,9 +1308,12 @@ class ABIArg
     explicit ABIArg(FloatRegister fpu) : kind_(FPU) { u.fpu_ = fpu.code(); }
     explicit ABIArg(uint32_t offset) : kind_(Stack) { u.offset_ = offset; }
 
-    Kind kind() const { return kind_; }
+    Kind kind() const {
+        MOZ_ASSERT(kind_ != Uninitialized);
+        return kind_;
+    }
 #ifdef JS_CODEGEN_REGISTER_PAIR
-    bool isGeneralRegPair() const { return kind_ == GPR_PAIR; }
+    bool isGeneralRegPair() const { return kind() == GPR_PAIR; }
 #else
     bool isGeneralRegPair() const { return false; }
 #endif
@@ -1274,26 +1337,32 @@ class ABIArg
         MOZ_ASSERT(isGeneralRegPair());
         return Register::FromCode(u.gpr_ + 1);
     }
-    FloatRegister fpu() const { MOZ_ASSERT(kind() == FPU); return FloatRegister::FromCode(u.fpu_); }
-    uint32_t offsetFromArgBase() const { MOZ_ASSERT(kind() == Stack); return u.offset_; }
+    FloatRegister fpu() const {
+        MOZ_ASSERT(kind() == FPU);
+        return FloatRegister::FromCode(u.fpu_);
+    }
+    uint32_t offsetFromArgBase() const {
+        MOZ_ASSERT(kind() == Stack);
+        return u.offset_;
+    }
 
     bool argInRegister() const { return kind() != Stack; }
-    AnyRegister reg() const { return kind_ == GPR ? AnyRegister(gpr()) : AnyRegister(fpu()); }
+    AnyRegister reg() const { return kind() == GPR ? AnyRegister(gpr()) : AnyRegister(fpu()); }
 
     bool operator==(const ABIArg& rhs) const {
         if (kind_ != rhs.kind_)
             return false;
 
-        switch((int8_t)kind_) {
+        switch(kind_) {
             case GPR:   return u.gpr_ == rhs.u.gpr_;
 #if defined(JS_CODEGEN_REGISTER_PAIR)
             case GPR_PAIR: return u.gpr_ == rhs.u.gpr_;
 #endif
             case FPU:   return u.fpu_ == rhs.u.fpu_;
             case Stack: return u.offset_ == rhs.u.offset_;
-            case -1:    return true;
-            default:    MOZ_CRASH("Invalid value for ABIArg kind");
+            case Uninitialized: return true;
         }
+        MOZ_CRASH("Invalid value for ABIArg kind");
     }
 
     bool operator!=(const ABIArg& rhs) const {
@@ -1305,7 +1374,7 @@ class ABIArg
 // clobbers all registers besides |unused|, but does not clobber floating point
 // registers.
 inline LiveGeneralRegisterSet
-SavedNonVolatileRegisters(AllocatableGeneralRegisterSet unused)
+SavedNonVolatileRegisters(const AllocatableGeneralRegisterSet& unused)
 {
     LiveGeneralRegisterSet result;
 
