@@ -7,17 +7,18 @@
 #include "vm/TraceLogging.h"
 
 #include "mozilla/DebugOnly.h"
+#include "mozilla/EndianUtils.h"
+#include "mozilla/MemoryReporting.h"
 #include "mozilla/ScopeExit.h"
 
 #include <string.h>
 
 #include "jsapi.h"
-#include "jsprf.h"
-#include "jsscript.h"
 
 #include "jit/BaselineJIT.h"
 #include "jit/CompileWrappers.h"
 #include "threading/LockGuard.h"
+#include "vm/JSScript.h"
 #include "vm/Runtime.h"
 #include "vm/Time.h"
 #include "vm/TraceLoggingGraph.h"
@@ -28,7 +29,6 @@ using namespace js;
 using namespace js::jit;
 
 using mozilla::DebugOnly;
-using mozilla::NativeEndian;
 
 TraceLoggerThreadState* traceLoggerState = nullptr;
 
@@ -60,7 +60,7 @@ rdtsc(void)
     return result;
 
 }
-#elif defined(__arm__)
+#elif defined(__arm__) || defined(__aarch64__)
 
 #include <sys/time.h>
 
@@ -103,6 +103,12 @@ EnsureTraceLoggerState()
     return true;
 }
 
+size_t
+js::SizeOfTraceLogState(mozilla::MallocSizeOf mallocSizeOf)
+{
+    return traceLoggerState ? traceLoggerState->sizeOfIncludingThis(mallocSizeOf) : 0;
+}
+
 void
 js::DestroyTraceLoggerThreadState()
 {
@@ -112,21 +118,25 @@ js::DestroyTraceLoggerThreadState()
     }
 }
 
+#ifdef DEBUG
+bool
+js::CurrentThreadOwnsTraceLoggerThreadStateLock()
+{
+    return traceLoggerState && traceLoggerState->lock.ownedByCurrentThread();
+}
+#endif
+
 void
-js::DestroyTraceLoggerMainThread(JSRuntime* runtime)
+js::DestroyTraceLogger(TraceLoggerThread* logger)
 {
     if (!EnsureTraceLoggerState())
         return;
-    traceLoggerState->destroyMainThread(runtime);
+    traceLoggerState->destroyLogger(logger);
 }
 
 bool
 TraceLoggerThread::init()
 {
-    if (!pointerMap.init())
-        return false;
-    if (!textIdPayloads.init())
-        return false;
     if (!events.init())
         return false;
 
@@ -175,11 +185,6 @@ TraceLoggerThread::~TraceLoggerThread()
             graph->log(events);
         graph = nullptr;
     }
-
-    if (textIdPayloads.initialized()) {
-        for (TextIdHashMap::Range r = textIdPayloads.all(); !r.empty(); r.popFront())
-            js_delete(r.front().value());
-    }
 }
 
 bool
@@ -209,6 +214,33 @@ TraceLoggerThread::fail(JSContext* cx, const char* error)
     return false;
 }
 
+void
+TraceLoggerThread::silentFail(const char* error)
+{
+    traceLoggerState->maybeSpewError(error);
+    failed = true;
+    enabled_ = 0;
+}
+
+size_t
+TraceLoggerThread::sizeOfExcludingThis(mozilla::MallocSizeOf mallocSizeOf) const
+{
+    size_t size = 0;
+#ifdef DEBUG
+    size += graphStack.sizeOfExcludingThis(mallocSizeOf);
+#endif
+    size += events.sizeOfExcludingThis(mallocSizeOf);
+    if (graph.get())
+        size += graph->sizeOfIncludingThis(mallocSizeOf);
+    return size;
+}
+
+size_t
+TraceLoggerThread::sizeOfIncludingThis(mozilla::MallocSizeOf mallocSizeOf) const
+{
+    return mallocSizeOf(this) + sizeOfExcludingThis(mallocSizeOf);
+}
+
 bool
 TraceLoggerThread::enable(JSContext* cx)
 {
@@ -217,7 +249,7 @@ TraceLoggerThread::enable(JSContext* cx)
 
     if (enabled_ == 1) {
         // Get the top Activation to log the top script/pc (No inlined frames).
-        ActivationIterator iter(cx->runtime());
+        ActivationIterator iter(cx);
         Activation* act = iter.activation();
 
         if (!act)
@@ -227,20 +259,27 @@ TraceLoggerThread::enable(JSContext* cx)
         int32_t engine = 0;
 
         if (act->isJit()) {
-            JitFrameIterator it(iter);
+            JitFrameIter frame(iter->asJit());
 
-            while (!it.isScripted() && !it.done())
-                ++it;
+            while (!frame.done()) {
+                if (frame.isWasm()) {
+                    JS_ReportErrorNumberASCII(cx, GetErrorMessage, nullptr,
+                                              JSMSG_TRACELOGGER_ENABLE_FAIL,
+                                              "not yet supported in wasm code");
+                    return false;
+                }
+                if (frame.asJSJit().isScripted())
+                    break;
+                ++frame;
+            }
 
-            MOZ_ASSERT(!it.done());
-            MOZ_ASSERT(it.isIonJS() || it.isBaselineJS());
+            MOZ_ASSERT(!frame.done());
 
-            script = it.script();
-            engine = it.isIonJS() ? TraceLogger_IonMonkey : TraceLogger_Baseline;
-        } else if (act->isWasm()) {
-            JS_ReportErrorNumberASCII(cx, GetErrorMessage, nullptr, JSMSG_TRACELOGGER_ENABLE_FAIL,
-                                      "not yet supported in wasm code");
-            return false;
+            const JSJitFrameIter& jitFrame = frame.asJSJit();
+            MOZ_ASSERT(jitFrame.isIonJS() || jitFrame.isBaselineJS());
+
+            script = jitFrame.script();
+            engine = jitFrame.isIonJS() ? TraceLogger_IonMonkey : TraceLogger_Baseline;
         } else {
             MOZ_ASSERT(act->isInterpreter());
             InterpreterFrame* fp = act->asInterpreter()->current();
@@ -252,7 +291,7 @@ TraceLoggerThread::enable(JSContext* cx)
         if (script->compartment() != cx->compartment())
             return fail(cx, "compartment mismatch");
 
-        TraceLoggerEvent event(this, TraceLogger_Scripts, script);
+        TraceLoggerEvent event(TraceLogger_Scripts, script);
         startEvent(event);
         startEvent(engine);
     }
@@ -286,17 +325,42 @@ TraceLoggerThread::disable(bool force, const char* error)
 }
 
 const char*
-TraceLoggerThread::eventText(uint32_t id)
+TraceLoggerThread::maybeEventText(uint32_t id)
 {
     if (id < TraceLogger_Last)
         return TLTextIdString(static_cast<TraceLoggerTextId>(id));
+    return traceLoggerState->maybeEventText(id);
+}
+
+const char*
+TraceLoggerThreadState::maybeEventText(uint32_t id)
+{
+    LockGuard<Mutex> guard(lock);
 
     TextIdHashMap::Ptr p = textIdPayloads.lookup(id);
-    MOZ_ASSERT(p);
+    if (!p)
+        return nullptr;
 
     return p->value()->string();
 }
 
+size_t
+TraceLoggerThreadState::sizeOfExcludingThis(mozilla::MallocSizeOf mallocSizeOf)
+{
+    LockGuard<Mutex> guard(lock);
+
+    // Do not count threadLoggers since they are counted by JSContext::traceLogger.
+
+    size_t size = 0;
+    size += pointerMap.sizeOfExcludingThis(mallocSizeOf);
+    if (textIdPayloads.initialized()) {
+        size += textIdPayloads.sizeOfExcludingThis(mallocSizeOf);
+        for (TextIdHashMap::Range r = textIdPayloads.all(); !r.empty(); r.popFront())
+            r.front().value()->sizeOfIncludingThis(mallocSizeOf);
+    }
+    return size;
+}
+
 bool
 TraceLoggerThread::textIdIsScriptEvent(uint32_t id)
 {
@@ -343,49 +407,24 @@ TraceLoggerThread::extractScriptDetails(uint32_t textId, const char** filename,
 }
 
 TraceLoggerEventPayload*
-TraceLoggerThread::getOrCreateEventPayload(TraceLoggerTextId textId)
+TraceLoggerThreadState::getOrCreateEventPayload(const char* text)
 {
-    TextIdHashMap::AddPtr p = textIdPayloads.lookupForAdd(textId);
-    if (p) {
-        MOZ_ASSERT(p->value()->textId() == textId); // Sanity check.
-        return p->value();
-    }
+    LockGuard<Mutex> guard(lock);
 
-    TraceLoggerEventPayload* payload = js_new<TraceLoggerEventPayload>(textId, (char*)nullptr);
-    if (!payload)
-        return nullptr;
-
-    if (!textIdPayloads.add(p, textId, payload))
-        return nullptr;
-
-    return payload;
-}
-
-TraceLoggerEventPayload*
-TraceLoggerThread::getOrCreateEventPayload(const char* text)
-{
     PointerHashMap::AddPtr p = pointerMap.lookupForAdd((const void*)text);
     if (p) {
         MOZ_ASSERT(p->value()->textId() < nextTextId); // Sanity check.
+        p->value()->use();
         return p->value();
     }
 
-    TraceLoggerEventPayload* payload = nullptr;
-
-    startEvent(TraceLogger_Internal);
-    auto guardInternalStopEvent = mozilla::MakeScopeExit([&] {
-        stopEvent(TraceLogger_Internal);
-        if (payload)
-            payload->release();
-    });
-
     char* str = js_strdup(text);
     if (!str)
         return nullptr;
 
     uint32_t textId = nextTextId;
 
-    payload = js_new<TraceLoggerEventPayload>(textId, str);
+    TraceLoggerEventPayload* payload = js_new<TraceLoggerEventPayload>(textId, str);
     if (!payload) {
         js_free(str);
         return nullptr;
@@ -397,54 +436,37 @@ TraceLoggerThread::getOrCreateEventPayload(const char* text)
         return nullptr;
     }
 
-    // Temporarily mark the payload as used. To make sure it doesn't get GC'ed.
     payload->use();
 
-    if (graph.get())
-        graph->addTextId(textId, str);
-
     nextTextId++;
 
     if (!pointerMap.add(p, text, payload))
         return nullptr;
 
+    payload->incPointerCount();
+
     return payload;
 }
 
 TraceLoggerEventPayload*
-TraceLoggerThread::getOrCreateEventPayload(TraceLoggerTextId type, const char* filename,
-                                           size_t lineno, size_t colno, const void* ptr)
+TraceLoggerThreadState::getOrCreateEventPayload(const char* filename,
+                                                size_t lineno, size_t colno, const void* ptr)
 {
-    MOZ_ASSERT(type == TraceLogger_Scripts || type == TraceLogger_AnnotateScripts ||
-               type == TraceLogger_InlinedScripts);
-
     if (!filename)
         filename = "<unknown>";
 
-    // Only log scripts when enabled otherwise return the global Scripts textId,
-    // which will get filtered out.
-    MOZ_ASSERT(traceLoggerState);
-    if (!traceLoggerState->isTextIdEnabled(type))
-        return getOrCreateEventPayload(type);
+    LockGuard<Mutex> guard(lock);
 
     PointerHashMap::AddPtr p;
     if (ptr) {
         p = pointerMap.lookupForAdd(ptr);
         if (p) {
             MOZ_ASSERT(p->value()->textId() < nextTextId); // Sanity check.
+            p->value()->use();
             return p->value();
         }
     }
 
-    TraceLoggerEventPayload* payload = nullptr;
-
-    startEvent(TraceLogger_Internal);
-    auto guardInternalStopEvent = mozilla::MakeScopeExit([&] {
-        stopEvent(TraceLogger_Internal);
-        if (payload)
-            payload->release();
-    });
-
     // Compute the length of the string to create.
     size_t lenFilename = strlen(filename);
     size_t lenLineno = 1;
@@ -458,12 +480,12 @@ TraceLoggerThread::getOrCreateEventPayload(TraceLoggerTextId type, const char* f
         return nullptr;
 
     DebugOnly<size_t> ret =
-        snprintf(str, len + 1, "script %s:%" PRIuSIZE ":%" PRIuSIZE, filename, lineno, colno);
+        snprintf(str, len + 1, "script %s:%zu:%zu", filename, lineno, colno);
     MOZ_ASSERT(ret == len);
     MOZ_ASSERT(strlen(str) == len);
 
     uint32_t textId = nextTextId;
-    payload = js_new<TraceLoggerEventPayload>(textId, str);
+    TraceLoggerEventPayload* payload = js_new<TraceLoggerEventPayload>(textId, str);
     if (!payload) {
         js_free(str);
         return nullptr;
@@ -475,34 +497,50 @@ TraceLoggerThread::getOrCreateEventPayload(TraceLoggerTextId type, const char* f
         return nullptr;
     }
 
-    // Temporarily mark the payload as used. To make sure it doesn't get GC'ed.
     payload->use();
 
-    if (graph.get())
-        graph->addTextId(textId, str);
-
     nextTextId++;
 
     if (ptr) {
         if (!pointerMap.add(p, ptr, payload))
             return nullptr;
+
+        payload->incPointerCount();
     }
 
     return payload;
 }
 
 TraceLoggerEventPayload*
-TraceLoggerThread::getOrCreateEventPayload(TraceLoggerTextId type, JSScript* script)
+TraceLoggerThreadState::getOrCreateEventPayload(JSScript* script)
 {
-    return getOrCreateEventPayload(type, script->filename(), script->lineno(), script->column(),
-                                   nullptr);
+    return getOrCreateEventPayload(script->filename(), script->lineno(), script->column(), nullptr);
 }
 
-TraceLoggerEventPayload*
-TraceLoggerThread::getOrCreateEventPayload(TraceLoggerTextId type,
-                                           const JS::ReadOnlyCompileOptions& script)
+void
+TraceLoggerThreadState::purgeUnusedPayloads()
 {
-    return getOrCreateEventPayload(type, script.filename(), script.lineno, script.column, nullptr);
+    // Care needs to be taken to maintain a coherent state in this function,
+    // as payloads can have their use count change at any time from non-zero to
+    // zero (but not the other way around; see TraceLoggerEventPayload::use()).
+    LockGuard<Mutex> guard(lock);
+
+    // Remove all the pointers to payloads that have no uses anymore
+    // and decrease the pointer count of that payload.
+    for (PointerHashMap::Enum e(pointerMap); !e.empty(); e.popFront()) {
+        if (e.front().value()->uses() == 0) {
+            e.front().value()->decPointerCount();
+            e.removeFront();
+        }
+    }
+
+    // Free all other payloads that have no uses anymore.
+    for (TextIdHashMap::Enum e(textIdPayloads); !e.empty(); e.popFront()) {
+        if (e.front().value()->uses() == 0 && e.front().value()->pointerCount() == 0) {
+            js_delete(e.front().value());
+            e.removeFront();
+        }
+    }
 }
 
 void
@@ -512,7 +550,7 @@ TraceLoggerThread::startEvent(TraceLoggerTextId id) {
 
 void
 TraceLoggerThread::startEvent(const TraceLoggerEvent& event) {
-    if (!event.hasPayload()) {
+    if (!event.hasTextId()) {
         if (!enabled())
             return;
         startEvent(TraceLogger_Error);
@@ -521,7 +559,7 @@ TraceLoggerThread::startEvent(const TraceLoggerEvent& event) {
                                     "this event. Disabling TraceLogger.");
         return;
     }
-    startEvent(event.payload()->textId());
+    startEvent(event.textId());
 }
 
 void
@@ -540,6 +578,11 @@ TraceLoggerThread::startEvent(uint32_t id)
     }
 #endif
 
+    if (graph.get()) {
+        for (uint32_t otherId = graph->nextTextId(); otherId <= id; otherId++)
+            graph->addTextId(otherId, maybeEventText(id));
+    }
+
     log(id);
 }
 
@@ -550,11 +593,11 @@ TraceLoggerThread::stopEvent(TraceLoggerTextId id) {
 
 void
 TraceLoggerThread::stopEvent(const TraceLoggerEvent& event) {
-    if (!event.hasPayload()) {
+    if (!event.hasTextId()) {
         stopEvent(TraceLogger_Error);
         return;
     }
-    stopEvent(event.payload()->textId());
+    stopEvent(event.textId());
 }
 
 void
@@ -578,7 +621,10 @@ TraceLoggerThread::stopEvent(uint32_t id)
             MOZ_ASSERT(prev >= TraceLogger_Last);
         } else if (id >= TraceLogger_Last) {
             MOZ_ASSERT(prev >= TraceLogger_Last);
-            MOZ_ASSERT_IF(prev != id, strcmp(eventText(id), eventText(prev)) == 0);
+            if (prev != id) {
+                // Ignore if the text has been flushed already.
+                MOZ_ASSERT_IF(maybeEventText(prev), strcmp(eventText(id), eventText(prev)) == 0);
+            }
         } else {
             MOZ_ASSERT(id == prev);
         }
@@ -627,26 +673,8 @@ TraceLoggerThread::log(uint32_t id)
             iteration_++;
             events.clear();
 
-            // Remove the item in the pointerMap for which the payloads
-            // have no uses anymore
-            for (PointerHashMap::Enum e(pointerMap); !e.empty(); e.popFront()) {
-                if (e.front().value()->uses() != 0)
-                    continue;
-
-                TextIdHashMap::Ptr p = textIdPayloads.lookup(e.front().value()->textId());
-                MOZ_ASSERT(p);
-                textIdPayloads.remove(p);
-
-                e.removeFront();
-            }
-
-            // Free all payloads that have no uses anymore.
-            for (TextIdHashMap::Enum e(textIdPayloads); !e.empty(); e.popFront()) {
-                if (e.front().value()->uses() == 0) {
-                    js_delete(e.front().value());
-                    e.removeFront();
-                }
-            }
+            // Periodically remove unused payloads from the global logger state.
+            traceLoggerState->purgeUnusedPayloads();
         }
 
         // Log the time it took to flush the events as being from the
@@ -673,14 +701,14 @@ TraceLoggerThread::log(uint32_t id)
 
 TraceLoggerThreadState::~TraceLoggerThreadState()
 {
-    while (TraceLoggerMainThread* logger = traceLoggerMainThreadList.popFirst())
+    while (TraceLoggerThread* logger = threadLoggers.popFirst())
         js_delete(logger);
 
-    if (threadLoggers.initialized()) {
-        for (ThreadLoggerHashMap::Range r = threadLoggers.all(); !r.empty(); r.popFront())
-            js_delete(r.front().value());
+    threadLoggers.clear();
 
-        threadLoggers.finish();
+    if (textIdPayloads.initialized()) {
+        for (TextIdHashMap::Range r = textIdPayloads.all(); !r.empty(); r.popFront())
+            js_delete(r.front().value());
     }
 
 #ifdef DEBUG
@@ -704,9 +732,6 @@ ContainsFlag(const char* str, const char* flag)
 bool
 TraceLoggerThreadState::init()
 {
-    if (!threadLoggers.init())
-        return false;
-
     const char* env = getenv("TLLOG");
     if (!env)
         env = "";
@@ -721,9 +746,9 @@ TraceLoggerThreadState::init()
             "  Default        Output all default. It includes:\n"
             "                 AnnotateScripts, Bailout, Baseline, BaselineCompilation, GC,\n"
             "                 GCAllocation, GCSweeping, Interpreter, IonAnalysis, IonCompilation,\n"
-            "                 IonLinking, IonMonkey, MinorGC, ParserCompileFunction,\n"
-            "                 ParserCompileScript, ParserCompileLazy, ParserCompileModule,\n"
-            "                 IrregexpCompile, IrregexpExecute, Scripts, Engine, WasmCompilation\n"
+            "                 IonLinking, IonMonkey, MinorGC, Frontend, ParsingFull,\n"
+            "                 ParsingSyntax, BytecodeEmission, IrregexpCompile, IrregexpExecute,\n"
+            "                 Scripts, Engine, WasmCompilation\n"
             "\n"
             "  IonCompiler    Output all information about compilation. It includes:\n"
             "                 IonCompilation, IonLinking, PruneUnusedBranches, FoldTests,\n"
@@ -736,8 +761,11 @@ TraceLoggerThreadState::init()
             "                 AddKeepAliveInstructions, GenerateLIR, RegisterAllocation, \n"
             "                 GenerateCode, Scripts, IonBuilderRestartLoop\n"
             "\n"
-            "  VMSpecific     Output the specific name of the VM call"
+            "  VMSpecific     Output the specific name of the VM call\n"
             "\n"
+            "  Frontend       Output all information about frontend compilation. It includes:\n"
+            "                 Frontend, ParsingFull, ParsingSyntax, Tokenizing,\n"
+            "                 BytecodeEmission, BytecodeFoldConstants, BytecodeNameFunctions\n"
             "Specific log items:\n"
         );
         for (uint32_t i = 1; i < TraceLogger_Last; i++) {
@@ -773,10 +801,10 @@ TraceLoggerThreadState::init()
         enabledTextIds[TraceLogger_IonLinking] = true;
         enabledTextIds[TraceLogger_IonMonkey] = true;
         enabledTextIds[TraceLogger_MinorGC] = true;
-        enabledTextIds[TraceLogger_ParserCompileFunction] = true;
-        enabledTextIds[TraceLogger_ParserCompileLazy] = true;
-        enabledTextIds[TraceLogger_ParserCompileScript] = true;
-        enabledTextIds[TraceLogger_ParserCompileModule] = true;
+        enabledTextIds[TraceLogger_Frontend] = true;
+        enabledTextIds[TraceLogger_ParsingFull] = true;
+        enabledTextIds[TraceLogger_ParsingSyntax] = true;
+        enabledTextIds[TraceLogger_BytecodeEmission] = true;
         enabledTextIds[TraceLogger_IrregexpCompile] = true;
         enabledTextIds[TraceLogger_IrregexpExecute] = true;
         enabledTextIds[TraceLogger_Scripts] = true;
@@ -818,6 +846,15 @@ TraceLoggerThreadState::init()
         enabledTextIds[TraceLogger_IonBuilderRestartLoop] = true;
     }
 
+    if (ContainsFlag(env, "Frontend")) {
+        enabledTextIds[TraceLogger_Frontend] = true;
+        enabledTextIds[TraceLogger_ParsingFull] = true;
+        enabledTextIds[TraceLogger_ParsingSyntax] = true;
+        enabledTextIds[TraceLogger_BytecodeEmission] = true;
+        enabledTextIds[TraceLogger_BytecodeFoldConstants] = true;
+        enabledTextIds[TraceLogger_BytecodeNameFunctions] = true;
+    }
+
     enabledTextIds[TraceLogger_Interpreter] = enabledTextIds[TraceLogger_Engine];
     enabledTextIds[TraceLogger_Baseline] = enabledTextIds[TraceLogger_Engine];
     enabledTextIds[TraceLogger_IonMonkey] = enabledTextIds[TraceLogger_Engine];
@@ -832,7 +869,7 @@ TraceLoggerThreadState::init()
                 "\n"
                 "usage: TLOPTIONS=option,option,option,... where options can be:\n"
                 "\n"
-                "  EnableMainThread        Start logging the main thread immediately.\n"
+                "  EnableActiveThread      Start logging cooperating threads immediately.\n"
                 "  EnableOffThread         Start logging helper threads immediately.\n"
                 "  EnableGraph             Enable spewing the tracelogging graph to a file.\n"
                 "  Errors                  Report errors during tracing to stderr.\n"
@@ -842,16 +879,21 @@ TraceLoggerThreadState::init()
             /*NOTREACHED*/
         }
 
-        if (strstr(options, "EnableMainThread"))
-            mainThreadEnabled = true;
+        if (strstr(options, "EnableActiveThread"))
+            cooperatingThreadEnabled = true;
         if (strstr(options, "EnableOffThread"))
-            offThreadEnabled = true;
+            helperThreadEnabled = true;
         if (strstr(options, "EnableGraph"))
             graphSpewingEnabled = true;
         if (strstr(options, "Errors"))
             spewErrors = true;
     }
 
+    if (!pointerMap.init())
+        return false;
+    if (!textIdPayloads.init())
+        return false;
+
     startupTime = rdtsc();
 
 #ifdef DEBUG
@@ -907,43 +949,28 @@ TraceLoggerThreadState::disableTextId(JSContext* cx, uint32_t textId)
         jit::ToggleBaselineTraceLoggerEngine(cx->runtime(), false);
 }
 
-
 TraceLoggerThread*
-js::TraceLoggerForMainThread(CompileRuntime* runtime)
+js::TraceLoggerForCurrentThread(JSContext* maybecx)
 {
     if (!EnsureTraceLoggerState())
         return nullptr;
-    return traceLoggerState->forMainThread(runtime);
+    return traceLoggerState->forCurrentThread(maybecx);
 }
 
 TraceLoggerThread*
-TraceLoggerThreadState::forMainThread(CompileRuntime* runtime)
-{
-    return forMainThread(runtime->mainThread());
-}
-
-TraceLoggerThread*
-js::TraceLoggerForMainThread(JSRuntime* runtime)
-{
-    if (!EnsureTraceLoggerState())
-        return nullptr;
-    return traceLoggerState->forMainThread(runtime);
-}
-
-TraceLoggerThread*
-TraceLoggerThreadState::forMainThread(JSRuntime* runtime)
-{
-    return forMainThread(&runtime->mainThread);
-}
-
-TraceLoggerThread*
-TraceLoggerThreadState::forMainThread(PerThreadData* mainThread)
+TraceLoggerThreadState::forCurrentThread(JSContext* maybecx)
 {
     MOZ_ASSERT(initialized);
-    if (!mainThread->traceLogger) {
+    MOZ_ASSERT_IF(maybecx, maybecx == TlsContext.get());
+
+    JSContext* cx = maybecx ? maybecx : TlsContext.get();
+    if (!cx)
+        return nullptr;
+
+    if (!cx->traceLogger) {
         LockGuard<Mutex> guard(lock);
 
-        TraceLoggerMainThread* logger = js_new<TraceLoggerMainThread>();
+        TraceLoggerThread* logger = js_new<TraceLoggerThread>();
         if (!logger)
             return nullptr;
 
@@ -952,45 +979,28 @@ TraceLoggerThreadState::forMainThread(PerThreadData* mainThread)
             return nullptr;
         }
 
-        traceLoggerMainThreadList.insertFront(logger);
-        mainThread->traceLogger = logger;
+        threadLoggers.insertFront(logger);
+        cx->traceLogger = logger;
 
         if (graphSpewingEnabled)
             logger->initGraph();
 
-        if (mainThreadEnabled)
+        if (CurrentHelperThread() ? helperThreadEnabled : cooperatingThreadEnabled)
             logger->enable();
     }
 
-    return mainThread->traceLogger;
+    return cx->traceLogger;
 }
 
 void
-TraceLoggerThreadState::destroyMainThread(JSRuntime* runtime)
+TraceLoggerThreadState::destroyLogger(TraceLoggerThread* logger)
 {
     MOZ_ASSERT(initialized);
-    PerThreadData* mainThread = &runtime->mainThread;
-    if (mainThread->traceLogger) {
-        LockGuard<Mutex> guard(lock);
+    MOZ_ASSERT(logger);
+    LockGuard<Mutex> guard(lock);
 
-        mainThread->traceLogger->remove();
-        js_delete(mainThread->traceLogger);
-        mainThread->traceLogger = nullptr;
-    }
-}
-
-TraceLoggerThread*
-js::TraceLoggerForCurrentThread()
-{
-    if (!EnsureTraceLoggerState())
-        return nullptr;
-    return traceLoggerState->forThread(ThisThread::GetId());
-}
-
-TraceLoggerThread*
-TraceLoggerThreadState::forThread(const Thread::Id& thread)
-{
-    return nullptr;
+    logger->remove();
+    js_delete(logger);
 }
 
 bool
@@ -1016,61 +1026,60 @@ js::TraceLogDisableTextId(JSContext* cx, uint32_t textId)
     traceLoggerState->disableTextId(cx, textId);
 }
 
-TraceLoggerEvent::TraceLoggerEvent(TraceLoggerThread* logger, TraceLoggerTextId textId)
+TraceLoggerEvent::TraceLoggerEvent(TraceLoggerTextId type, JSScript* script)
+  : TraceLoggerEvent(type, script->filename(), script->lineno(), script->column())
+{ }
+
+TraceLoggerEvent::TraceLoggerEvent(TraceLoggerTextId type, const char* filename, size_t line,
+                                   size_t column)
+  : payload_()
 {
-    payload_ = nullptr;
-    if (logger) {
-        payload_ = logger->getOrCreateEventPayload(textId);
-        if (payload_)
-            payload_->use();
+    MOZ_ASSERT(type == TraceLogger_Scripts || type == TraceLogger_AnnotateScripts ||
+               type == TraceLogger_InlinedScripts || type == TraceLogger_Frontend);
+
+    if (!traceLoggerState)
+        return;
+
+    // Only log scripts when enabled, otherwise use the more generic type
+    // (which will get filtered out).
+    if (!traceLoggerState->isTextIdEnabled(type)) {
+        payload_.setTextId(type);
+        return;
     }
+
+    payload_.setEventPayload(
+        traceLoggerState->getOrCreateEventPayload(filename, line, column, nullptr));
 }
 
-TraceLoggerEvent::TraceLoggerEvent(TraceLoggerThread* logger, TraceLoggerTextId type,
-                                   JSScript* script)
+TraceLoggerEvent::TraceLoggerEvent(const char* text)
+  : payload_()
 {
-    payload_ = nullptr;
-    if (logger) {
-        payload_ = logger->getOrCreateEventPayload(type, script);
-        if (payload_)
-            payload_->use();
-    }
-}
-
-TraceLoggerEvent::TraceLoggerEvent(TraceLoggerThread* logger, TraceLoggerTextId type,
-                                   const JS::ReadOnlyCompileOptions& compileOptions)
-{
-    payload_ = nullptr;
-    if (logger) {
-        payload_ = logger->getOrCreateEventPayload(type, compileOptions);
-        if (payload_)
-            payload_->use();
-    }
-}
-
-TraceLoggerEvent::TraceLoggerEvent(TraceLoggerThread* logger, const char* text)
-{
-    payload_ = nullptr;
-    if (logger) {
-        payload_ = logger->getOrCreateEventPayload(text);
-        if (payload_)
-            payload_->use();
-    }
+    if (traceLoggerState)
+        payload_.setEventPayload(traceLoggerState->getOrCreateEventPayload(text));
 }
 
 TraceLoggerEvent::~TraceLoggerEvent()
 {
-    if (payload_)
-        payload_->release();
+    if (hasExtPayload())
+        extPayload()->release();
+}
+
+uint32_t
+TraceLoggerEvent::textId() const
+{
+    MOZ_ASSERT(hasTextId());
+    if (hasExtPayload())
+        return extPayload()->textId();
+    return payload_.textId();
 }
 
 TraceLoggerEvent&
 TraceLoggerEvent::operator=(const TraceLoggerEvent& other)
 {
-    if (other.hasPayload())
-        other.payload()->use();
-    if (hasPayload())
-        payload()->release();
+    if (other.hasExtPayload())
+        other.extPayload()->use();
+    if (hasExtPayload())
+        extPayload()->release();
 
     payload_ = other.payload_;
 
@@ -1078,8 +1087,8 @@ TraceLoggerEvent::operator=(const TraceLoggerEvent& other)
 }
 
 TraceLoggerEvent::TraceLoggerEvent(const TraceLoggerEvent& other)
+  : payload_(other.payload_)
 {
-    payload_ = other.payload_;
-    if (hasPayload())
-        payload()->use();
+    if (hasExtPayload())
+        extPayload()->use();
 }
